namespace ActiveAllocator.API.Controllers;

using AutoMapper;
using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.Options;
using ActiveAllocator.API.Models.Transactions;
using ActiveAllocator.API.Services;
using ActiveAllocator.API.Config;
using System.Runtime.InteropServices;
using ActiveAllocator.API.Entities;
using System.Collections.Generic;
using Microsoft.AspNetCore.Authorization;
using ActiveAllocator.API.Models.Transactions.VendorSpecific;
using ActiveAllocator.API.Helpers;
using Newtonsoft.Json;

[Authorize]
[ApiController]
[Route("[controller]")]
public class TransactionsController : ControllerBase
{
    private ITransactionService _transactionService;
    private IMapper _mapper;
    private readonly AppSettings _appSettings;
    private readonly ILogger<TransactionsController> _logger;

    public TransactionsController(
        ITransactionService transactionService,
        IMapper mapper,
        IOptions<AppSettings> appSettings,
        ILogger<TransactionsController> logger)
    {
        _transactionService = transactionService;
        _mapper = mapper;
        _appSettings = appSettings.Value;
        _logger = logger;
    }

    [HttpGet]
    public IActionResult GetAll([FromQuery]PaginationParameters paginationParameters, int? accountId = null)
    {
        List<TransactionDto> transactionDtos = new List<TransactionDto>();

        PagedList<Transaction> transactions = _transactionService.GetTransactions(paginationParameters, accountId);

        foreach (Transaction tran in transactions)
        {
            if (accountId.HasValue
            && (tran.DebitAccount == null || tran.DebitAccount.Id != accountId)
            && (tran.CreditAccount == null || tran.CreditAccount.Id != accountId))
                continue;

            transactionDtos.Add(_mapper.Map<Transaction, TransactionDto>(tran));
        }

        // Sort by Date
        //transactionDtos.Sort((t1, t2) => t2.Date.CompareTo(t1.Date));

        var metadata = new
        {
            transactions.TotalCount,
            transactions.PageSize,
            transactions.CurrentPage,
            transactions.TotalPages,
            transactions.HasNext,
            transactions.HasPrevious
        };

        Response.Headers.Add("X-Pagination", JsonConvert.SerializeObject(metadata));

        _logger.LogInformation($"Returned {transactions.PageSize} transactions from database.");
        return Ok(transactionDtos);
    }

    [HttpGet("{id}")]
    public IActionResult GetById(int id)
    {
        Transaction tran = _transactionService.GetById(id);
        return Ok(_mapper.Map<Transaction, TransactionDto>(tran));
    }

    [HttpPost("BulkAdd")]
    public IActionResult BulkCreate([FromBody]List<TransactionCreate> model)
    {
        List<Transaction> trans = _transactionService.BulkCreate(model).ToList();
        return Ok(new { message = $"{trans.Count()} transaction(s) created." });
    }

    [HttpPost("BulkAdd/SimmonsBank")]
    public IActionResult BulkCreateSimmonsBank(int debitAccountId, [FromBody]List<SimmonsBankTransactionCreate> model)
    {
        List<TransactionCreate> transactionCreates = new();

        foreach (SimmonsBankTransactionCreate sbTran in model)
        {
            decimal amount = Convert.ToDecimal(sbTran.Amount);

            transactionCreates.Add(new() {
                Date = sbTran.Date,
                ExternalId = sbTran.Id,
                Description = sbTran.Memo,
                DebitAccount = amount < 0 ? debitAccountId : 0,
                CreditAccount = amount >= 0 ? debitAccountId : 0,
                IsEnvelopeFundingTransaction = false,
                Amount = Math.Abs(amount),
                CurrencyType = 1,
                Notes = "SimmonsBank BulkAdd. ACCOUNT=" + sbTran.AccountId,
                IsPending = sbTran.PendingStatus != "NoReconciliation",
                Tags = new() {
                    "RunningBalance=" + sbTran.RunningBalance
                },
                TriggerAutoclassRules = true
            });
        }

        List<Transaction> trans = _transactionService.BulkCreate(transactionCreates).ToList();
        return Ok(new { message = $"{trans.Count} transaction(s) created." });
    }

    [HttpPost]
    public IActionResult Create([FromBody]TransactionCreate model)
    {
        Transaction tran = _transactionService.Create(model);
        return Ok(new { message = $"transaction '{tran.Description}' created with id '{tran.Id}'." });
    }

    [HttpPut("{id}")]
    public IActionResult Update(int id, [FromBody]TransactionUpdateRequest model)
    {
        _transactionService.Update(id, model);
        return Ok(new { message = $"transaction with id '{id}' updated" });
    }

    [HttpDelete("{id}")]
    public IActionResult Delete(int id)
    {
        _transactionService.Delete(id);
        return Ok(new { message = "transaction deleted" });
    }
}